/***************************************************************************
                          HorizontalOperations.h  -  description
                             -------------------
    begin                : Apr 18, 2019
    copyright            : (C) 2019 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <TNL/Math.h>

namespace TNL {
namespace Containers {
namespace Expressions {

template< typename T1, typename T2 > // Proč nejsou šablonové argumenty až u funkce evaluate (mohly by být automaticky odvozeny)
struct Addition
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( a + b )
   {
      return a + b;
   }
/*
   __cuda_callable__
   static auto evaluate( const T1& a, T2&& b ) -> decltype( a + b )
   {
      return a + std::forward<T2>(b);
   }

   __cuda_callable__
   static auto evaluate( T1&& a, const T2& b ) -> decltype( a + b )
   {
      return std::forward<T1>(a) + b;
   }


   __cuda_callable__
   static auto evaluate( T1&& a, T2&& b ) -> decltype( a + b )
   {
      return std::forward<T1>(a) + std::forward<T1>(b);
   }
*/
};

template< typename T1, typename T2 >
struct Subtraction
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( a - b )
   {
      return a - b;
   }
/*
   __cuda_callable__
   static auto evaluate( const T1& a, T2&& b ) -> decltype( a - b )
   {
      return a - std::forward<T2>(b);
   }

   __cuda_callable__
   static auto evaluate( T1&& a, const T2& b ) -> decltype( a - b )
   {
      return std::forward<T1>(a) - b;
   }


   __cuda_callable__
   static auto evaluate( T1&& a, T2&& b ) -> decltype( a - b )
   {
      return std::forward<T1>(a) - std::forward<T1>(b);
   }
*/
};

template< typename T1, typename T2 >
struct Multiplication
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( a * b )
   {
      return a * b;
   }
/*
   __cuda_callable__
   static auto evaluate( const T1& a, T2&& b ) -> decltype( a * b )
   {
      return a * std::forward<T2>(b);
   }

   __cuda_callable__
   static auto evaluate( T1&& a, const T2& b ) -> decltype( a * b )
   {
      return std::forward<T1>(a) * b;
   }


   __cuda_callable__
   static auto evaluate( T1&& a, T2&& b ) -> decltype( a * b )
   {
      return std::forward<T1>(a) * std::forward<T1>(b);
   }
*/
};

template< typename T1, typename T2 >
struct Division
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( a / b )
   {
      return a / b;
   }
/*
   __cuda_callable__
   static auto evaluate( const T1& a, T2&& b ) -> decltype( a / b )
   {
      return a / std::forward<T2>(b);
   }

   __cuda_callable__
   static auto evaluate( T1&& a, const T2& b ) -> decltype( a / b )
   {
      return std::forward<T1>(a) / b;
   }


   __cuda_callable__
   static auto evaluate( T1&& a, T2&& b ) -> decltype( a / b )
   {
      return std::forward<T1>(a) / std::forward<T1>(b);
   }
*/
};

template< typename T1, typename T2 >
struct Min
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( TNL::min( a , b ) )
   {
      return TNL::min( a, b );
   }
};

template< typename T1, typename T2 >
struct Max
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& b ) -> decltype( TNL::max( a, b ) )
   {
      return TNL::max( a, b );
   }
};

template< typename T1 >
struct Minus
{
   __cuda_callable__
   static T1 evaluate( const T1& a )
   {
      return -a;
   }
/*
   __cuda_callable__
   static T1 evaluate( T1&& a )
   {
      return -(std::forward<T1>(a));
   }
*/
};

template< typename T1 >
struct Abs
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> T1
   {
      using TNL::abs;
      return abs( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::abs( a ) )
   {
      return TNL::abs(std::forward<T1>(a));
   }*/
};

template< typename T1, typename T2 >
struct Pow
{
   __cuda_callable__
   static auto evaluate( const T1& a, const T2& exp ) -> decltype( TNL::pow( a, exp ) )
   {
      return TNL::pow( a, exp );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a, const T2& exp ) -> decltype( TNL::pow( a, exp ) )
   {
      return TNL::pow( std::forward<T1>(a), exp );
   }
*/
};

template< typename T1 >
struct Exp
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::exp( a ) )
   {
      return TNL::exp( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::exp( a ) )
   {
      return TNL::exp( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Sqrt
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::sqrt( a ) )
   {
      return TNL::sqrt( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::sqrt( a ) )
   {
      return TNL::sqrt( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Cbrt
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::cbrt( a ) )
   {
      return TNL::cbrt( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::cbrt( a ) )
   {
      return TNL::cbrt( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Log
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::log( a ) )
   {
      return TNL::log( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::log( a ) )
   {
      return TNL::log( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Log10
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::log10( a ) )
   {
      return TNL::log10( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::log10( a ) )
   {
      return TNL::log10( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Log2
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::log2( a ) )
   {
      return TNL::log2( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::log2( a ) )
   {
      return TNL::log2( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Sin
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::sin( a ) )
   {
      return TNL::sin( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::sin( a ) )
   {
      return TNL::sin( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Cos
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::cos( a ) )
   {
      return TNL::cos( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::cos( a ) )
   {
      return TNL::cos( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Tan
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::tan( a ) )
   {
      return TNL::tan( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::tan( a ) )
   {
      return TNL::tan( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Asin
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::asin( a ) )
   {
      return TNL::asin( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::asin( a ) )
   {
      return TNL::asin( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Acos
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::acos( a ) )
   {
      return TNL::acos( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::acos( a ) )
   {
      return TNL::acos( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Atan
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::atan( a ) )
   {
      return TNL::atan( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::atan( a ) )
   {
      return TNL::atan( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Sinh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::sinh( a ) )
   {
      return TNL::sinh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::sinh( a ) )
   {
      return TNL::sinh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Cosh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::cosh( a ) )
   {
      return TNL::cosh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::cosh( a ) )
   {
      return TNL::cosh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Tanh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::tanh( a ) )
   {
      return TNL::tanh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::tanh( a ) )
   {
      return TNL::tanh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Asinh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::asinh( a ) )
   {
      return TNL::asinh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::asinh( a ) )
   {
      return TNL::asinh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Acosh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::acosh( a ) )
   {
      return TNL::acosh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::acosh( a ) )
   {
      return TNL::acosh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Atanh
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::atanh( a ) )
   {
      return TNL::atanh( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::atanh( a ) )
   {
      return TNL::atanh( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Floor
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::floor( a ) )
   {
      return TNL::floor( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::floor( a ) )
   {
      return TNL::floor( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Ceil
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::ceil( a ) )
   {
      return TNL::ceil( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::ceil( a ) )
   {
      return TNL::ceil( std::forward<T1>(a) );
   }
*/
};

template< typename T1 >
struct Sign
{
   __cuda_callable__
   static auto evaluate( const T1& a ) -> decltype( TNL::sign( a ) )
   {
      return TNL::sign( a );
   }
/*
   __cuda_callable__
   static auto evaluate( T1&& a ) -> decltype( TNL::sign( a ) )
   {
      return TNL::sign( std::forward<T1>(a) );
   }
*/
};

template< typename ResultType >
struct Cast
{
   template< typename T1 >
   struct Operation
   {
      __cuda_callable__
      static auto evaluate( const T1& a ) -> ResultType
      {
         return static_cast<ResultType>( a );
      }
   };
};

} // namespace Expressions
} // namespace Containers
} // namespace TNL
